from datetime import datetime

from twisted.trial.unittest import TestCase

from go_metrics.metrics.graphite_time_parser import (
    TimeParserValueError, interval_to_seconds, parse_time)


class TestGraphiteTimeParser(TestCase):
    """
    Tests for Graphite-compatible time parsing functions.
    """

    def assert_TPVE(self, *args, **kw):
        """
        Convenience wrapper for asserting TimeParserValueError is raised.
        """
        return self.assertRaises(TimeParserValueError, *args, **kw)

    def assert_interval_to_seconds(self, expected, *inputs):
        """
        Convenience wrapper for testing multiple equivalent interval strings at
        once.
        """
        for value in inputs:
            self.assertEqual(interval_to_seconds(value), expected)

    def test_interval_to_seconds_with_invalid_value(self):
        """
        A TimeParserValueError is raised for various invalid values.
        """
        self.assert_TPVE(interval_to_seconds, "")
        self.assert_TPVE(interval_to_seconds, "s")
        self.assert_TPVE(interval_to_seconds, "1")
        self.assert_TPVE(interval_to_seconds, "three days")
        self.assert_TPVE(interval_to_seconds, "-1h")
        self.assert_TPVE(interval_to_seconds, "2fortnights")

    def test_interval_to_seconds_with_seconds(self):
        """
        With a suffix of "s" or "seconds", the interval is parsed in seconds.
        """
        self.assert_interval_to_seconds(0, "0s", "0second", "0seconds")
        self.assert_interval_to_seconds(1, "1s", "1second", "1seconds")
        self.assert_interval_to_seconds(60, "60s", "60second", "60seconds")
        self.assert_interval_to_seconds(
            1234567, "1234567s", "1234567second", "1234567seconds")
        self.assert_interval_to_seconds(12, "012s", "012second", "012seconds")

    def test_interval_to_seconds_with_minutes(self):
        """
        With a suffix of "min" or "minutes", the interval is parsed in minutes.
        """
        self.assert_interval_to_seconds(
            0, "0min", "0mins", "0minute", "0minutes")
        self.assert_interval_to_seconds(
            60, "1min", "1mins", "1minute", "1minutes")
        self.assert_interval_to_seconds(
            3600, "60min", "60mins", "60minute", "60minutes")
        self.assert_interval_to_seconds(
            1234567 * 60, "1234567min", "1234567mins", "1234567minute",
            "1234567minutes")
        self.assert_interval_to_seconds(
            720, "012min", "012mins", "012minute", "012minutes")

    def test_interval_to_seconds_with_hours(self):
        """
        With a suffix of "h" or "hours", the interval is parsed in hours.
        """
        self.assert_interval_to_seconds(0, "0h", "0hour", "0hours")
        self.assert_interval_to_seconds(3600, "1h", "1hour", "1hours")
        self.assert_interval_to_seconds(24 * 3600, "24h", "24hour", "24hours")
        self.assert_interval_to_seconds(
            1234 * 3600, "1234h", "1234hour", "1234hours")
        self.assert_interval_to_seconds(
            12 * 3600, "012h", "012hour", "012hours")

    def test_interval_to_seconds_with_days(self):
        """
        With a suffix of "d" or "days", the interval is parsed in days.
        """
        self.assert_interval_to_seconds(0, "0d", "0day", "0days")
        self.assert_interval_to_seconds(86400, "1d", "1day", "1days")
        self.assert_interval_to_seconds(60 * 86400, "60d", "60day", "60days")
        self.assert_interval_to_seconds(
            1234 * 86400, "1234d", "1234day", "1234days")
        self.assert_interval_to_seconds(
            14 * 86400, "014d", "014day", "014days")

    def test_interval_to_seconds_with_weeks(self):
        """
        With a suffix of "w" or "weeks", the interval is parsed in weeks.
        """
        self.assert_interval_to_seconds(0, "0w", "0week", "0weeks")
        self.assert_interval_to_seconds(604800, "1w", "1week", "1weeks")
        self.assert_interval_to_seconds(4 * 604800, "4w", "4week", "4weeks")
        self.assert_interval_to_seconds(
            123 * 604800, "123w", "123week", "123weeks")
        self.assert_interval_to_seconds(
            12 * 604800, "012w", "012week", "012weeks")

    def test_interval_to_seconds_with_months(self):
        """
        With a suffix of "mon" or "months", the interval is parsed in months.
        """
        self.assert_interval_to_seconds(0, "0mon", "0month", "0months")
        self.assert_interval_to_seconds(2592000, "1mon", "1month", "1months")
        self.assert_interval_to_seconds(
            12 * 2592000, "12mon", "12month", "12months")
        self.assert_interval_to_seconds(
            123 * 2592000, "123mon", "123month", "123months")
        self.assert_interval_to_seconds(
            12 * 2592000, "012mon", "012month", "012months")

    def test_interval_to_seconds_with_years(self):
        """
        With a suffix of "y" or "years", the interval is parsed in years.
        """
        self.assert_interval_to_seconds(0, "0y", "0year", "0years")
        self.assert_interval_to_seconds(31536000, "1y", "1year", "1years")
        self.assert_interval_to_seconds(5 * 31536000, "5y", "5year", "5years")
        self.assert_interval_to_seconds(
            123 * 31536000, "123y", "123year", "123years")
        self.assert_interval_to_seconds(
            2 * 31536000, "02y", "02year", "02years")

    def test_parse_time_with_invalid_interval(self):
        """
        If given a time_str starting with "-" and containing an invalid
        interval, a TimeParserValueError is raised.
        """
        now = datetime(2015, 2, 1, 0, 0, 0)
        self.assert_TPVE(parse_time, "-0", now)
        self.assert_TPVE(parse_time, "-12", now)
        self.assert_TPVE(parse_time, "-12fortnights", now)
        self.assert_TPVE(parse_time, "-20150101", now)

    def test_parse_time_with_interval(self):
        """
        If given a time_str starting with "-", the time is parsed as an
        interval prior to "now".
        """
        now = datetime(2015, 2, 1, 0, 0, 0)
        self.assertEqual(parse_time("-0s", now), now)
        self.assertEqual(
            parse_time("-1s", now), datetime(2015, 1, 31, 23, 59, 59))
        self.assertEqual(
            parse_time("-1s", now), datetime(2015, 1, 31, 23, 59, 59))
        self.assertEqual(
            parse_time("-2w", now), datetime(2015, 1, 18, 0, 0, 0))
        self.assertEqual(
            parse_time("-2w", datetime(2015, 1, 24, 10, 15, 25)),
            datetime(2015, 1, 10, 10, 15, 25))

    def test_parse_time_special_values(self):
        """
        The strings "now", "yesterday", "today", and "tomorrow" return what one
        might expect.
        """
        now1 = datetime(2015, 2, 1, 0, 0, 0)
        now2 = datetime(2015, 1, 24, 10, 15, 25)
        self.assertEqual(parse_time("now", now1), now1)
        self.assertEqual(parse_time("now", now2), now2)
        self.assertEqual(
            parse_time("yesterday", now1), datetime(2015, 1, 31, 0, 0, 0))
        self.assertEqual(
            parse_time("yesterday", now2), datetime(2015, 1, 23, 10, 15, 25))
        self.assertEqual(parse_time("today", now1), now1)
        self.assertEqual(parse_time("today", now2), now2)
        self.assertEqual(
            parse_time("tomorrow", now1), datetime(2015, 2, 2, 0, 0, 0))
        self.assertEqual(
            parse_time("tomorrow", now2), datetime(2015, 1, 25, 10, 15, 25))

    def test_parse_time_absolute_datetime(self):
        """
        Absolute timestamps in the form ``HH:MM_YYYYMMDD`` are parsed into
        appropriate datetimes with seconds set to zero.
        """
        self.assertEqual(
            parse_time("00:00_20150201", None), datetime(2015, 2, 1, 0, 0, 0))
        self.assertEqual(
            parse_time("00:00_19700101", None), datetime(1970, 1, 1, 0, 0, 0))

    def test_parse_time_absolute_date(self):
        """
        Absolute timestamps in the form ``YYYYMMDD`` are parsed into
        appropriate datetimes with the time set to midnight.
        """
        self.assertEqual(
            parse_time("20150201", None), datetime(2015, 2, 1, 0, 0, 0))
        self.assertEqual(
            parse_time("19700101", None), datetime(1970, 1, 1, 0, 0, 0))
        self.assertEqual(
            parse_time("19010101", None), datetime(1901, 1, 1, 0, 0, 0))
        self.assertEqual(
            parse_time("99991231", None), datetime(9999, 12, 31, 0, 0, 0))

    def test_parse_time_unix_timestamp(self):
        """
        Unix timestamps are accepted as absolute time specifications.
        """
        self.assertEqual(
            parse_time("1422748800", None), datetime(2015, 2, 1, 0, 0, 0))
        self.assertEqual(parse_time("0", None), datetime(1970, 1, 1, 0, 0, 0))
        # The following are treated as unix timestamps, not YYYYMMDD strings.
        self.assertEqual(
            parse_time("19000101", None), datetime(1970, 8, 8, 21, 48, 21))
        self.assertEqual(
            parse_time("20150132", None), datetime(1970, 8, 22, 5, 15, 32))
        self.assertEqual(
            parse_time("20151301", None), datetime(1970, 8, 22, 5, 35, 1))

    def test_parse_time_with_invalid_absolute_datetime(self):
        """
        If given a time_str containing an invalid value, a TimeParserValueError
        is raised.
        """
        self.assert_TPVE(parse_time, "", None)
        self.assert_TPVE(parse_time, "blahblah", None)
        # This is detected as a YYYYMMDD string, but it's invalid.
        self.assert_TPVE(parse_time, "20150231", None)

        # Graphite accepts the following, we don't.
        self.assert_TPVE(parse_time, "2015_02_01", None)
        self.assert_TPVE(parse_time, "12:35 20150201", None)
        self.assert_TPVE(parse_time, "12:3520150201", None)
        self.assert_TPVE(parse_time, "12/31/99", None)
        self.assert_TPVE(parse_time, "6pm today", None)
        self.assert_TPVE(parse_time, "noon tomorrow", None)
        self.assert_TPVE(parse_time, "january 1", None)
        self.assert_TPVE(parse_time, "monday", None)



from urlparse import parse_qs as _parse_qs

from twisted.internet.defer import maybeDeferred

from confmodel import Config
from confmodel.fields import ConfigDict

from go_api.cyclone.handlers import ApiApplication, BaseHandler

from go_metrics.metrics.base import MetricsBackendError, BadMetricsQueryError
from go_metrics.metrics.graphite import GraphiteBackend


def parse_qs(qs):
    return dict(
        (k, v[0] if len(v) == 1 else v)
        for (k, v) in _parse_qs(qs).iteritems())


class MetricsHandler(BaseHandler):
    def get(self):
        query = parse_qs(self.request.query)
        d = maybeDeferred(self.model.get, **query)
        d.addCallback(self.write_object)
        d.addErrback(self.catch_err, 400, BadMetricsQueryError)
        d.addErrback(self.catch_err, 500, MetricsBackendError)
        d.addErrback(self.raise_err, 500, "Failed to retrieve metrics.")
        return d

    def _assert_dict(self, d):
        if not isinstance(d, dict):
            raise BadMetricsQueryError(
                "Invalid query %r, should be dict, not %s" % (d, type(d)))

    def post(self):
        data = self.parse_json(self.request.body)
        d = maybeDeferred(self._assert_dict, data)
        d.addCallback(lambda _: self.model.fire(**data))
        d.addCallback(self.write_object)
        d.addErrback(self.catch_err, 400, BadMetricsQueryError)
        d.addErrback(self.raise_err, 500, "Failed to fire metrics.")
        return d


class MetricsApiConfig(Config):
    backend = ConfigDict("Config for metrics backend", default={})


class MetricsApi(ApiApplication):
    config_required = True
    backend_class = GraphiteBackend

    @property
    def models(self):
        return (('/metrics/', MetricsHandler, self.get_metrics_model),)

    def initialize(self, settings, config):
        config = MetricsApiConfig(config)
        self.backend = self.backend_class(config.backend)

    def get_metrics_model(self, owner_id):
        return self.backend.get_model(owner_id)



"""
Graphite backend for the metrics api.
"""

from datetime import datetime
from urllib import urlencode
from urlparse import urljoin

from twisted.internet.defer import (
    Deferred, inlineCallbacks, succeed, returnValue)

import treq

from confmodel.errors import ConfigError
from confmodel.fields import ConfigText, ConfigBool, ConfigInt

from vumi.blinkenlights.metrics import (
    Metric, MetricManager, SUM, AVG, MAX, MIN, LAST)
from vumi.service import Worker, WorkerCreator

from go_metrics.metrics.base import (
    Metrics, MetricsBackend, MetricsBackendError, BadMetricsQueryError)
from go_metrics.metrics.graphite_time_parser import (
    interval_to_seconds, parse_time)


def strip_aggregator(name, aggregator):
    name = name.split('.')
    if name[-1] == aggregator.name:
        return '.'.join(name[:-1])
    return '.'.join(name)


def is_error(resp):
    return 400 <= resp.code <= 599


def omit_nulls(datapoints):
    return [d for d in datapoints if d['y'] is not None]


def zeroize_nulls(datapoints):
    return [{
        'x': d['x'],
        'y': 0.0 if d['y'] is None else d['y']
    } for d in datapoints]


null_parsers = {
    'keep': lambda x: x,
    'omit': omit_nulls,
    'zeroize': zeroize_nulls,
}


class GraphiteMetrics(Metrics):
    aggregators = {
        'sum': SUM,
        'avg': AVG,
        'max': MAX,
        'min': MIN,
        'last': LAST,
    }

    def _agg_from_name(self, name):
        aggregator_name = name.split('.')[-1]
        aggregator = self.aggregators.get(aggregator_name)
        if aggregator is None:
            raise BadMetricsQueryError(
                "Aggregator '%s' is not a valid aggregator" % aggregator_name)
        return aggregator

    def _get_full_metric_name(self, name):
        return '%s.%s.%s' % (
            self.backend.config.prefix, self.owner_id, name)

    def _build_metric_name(self, name, interval, align_to_from):
        agg = self._agg_from_name(name).name
        full_name = self._get_full_metric_name(name)

        return (
            "alias(summarize(%s, '%s', '%s', %s), '%s')" %
            (full_name, interval, agg, align_to_from, name))

    def _build_render_url(self, params):
        metrics = params['m']

        targets = [
            self._build_metric_name(
                name, params['interval'], params['align_to_from'])
            for name in metrics]

        url = urljoin(self.backend.config.graphite_url, 'render/')
        return "%s?%s" % (url, urlencode({
            'format': 'json',
            'target': targets,
            'from': params['from'],
            'until': params['until'],
        }, True))

    def _parse_datapoints(self, datapoints):
        return [{
            'x': x * 1000,
            'y': y,
        } for (y, x) in datapoints]

    def _parse_response(self, data, null_parser):
        return dict(
            (d['target'], null_parser(self._parse_datapoints(d['datapoints'])))
            for d in data)

    def _get_auth(self):
        config = self.backend.config

        if config.username is not None and config.password is not None:
            return (config.username, config.password)
        else:
            return None

    def _predict_data_size(self, start, end, interval):
        """
        Use the start and end times and interval size to predict the number of
        data points being requested.
        """
        now = datetime.utcnow()
        # "end" can be earlier than "start".
        period = abs(parse_time(end, now) - parse_time(start, now))
        interval_secs = interval_to_seconds(interval)
        return (period.seconds + 86400 * period.days) / interval_secs

    @inlineCallbacks
    def get(self, **kw):
        params = {
            'm': [],
            'from': '-24h',
            'until': '-0s',
            'nulls': 'zeroize',
            'interval': '1hour',
            'align_to_from': 'false',
        }
        params.update(kw)

        if (isinstance(params['m'], basestring)):
            params['m'] = [params['m']]

        predicted_size = self._predict_data_size(
            params['from'], params['until'], params['interval'])
        predicted_size *= max(1, len(params['m']))
        max_response_size = self.backend.config.max_response_size
        if predicted_size > max_response_size:
            raise BadMetricsQueryError(
                "%s data points requested, maximum allowed is %s" % (
                    predicted_size, max_response_size))

        if params['nulls'] not in null_parsers:
            raise BadMetricsQueryError(
                "Unrecognised null parser '%s'" % (params['nulls'],))

        url = self._build_render_url(params)
        resp = yield treq.get(
            url,
            auth=self._get_auth(),
            persistent=self.backend.config.persistent)

        if is_error(resp):
            raise MetricsBackendError(
                "Got error response interacting with metrics backend")

        null_parser = null_parsers[params['nulls']]
        returnValue(self._parse_response((yield resp.json()), null_parser))

    @inlineCallbacks
    def fire(self, **kw):
        metrics = []
        metrics_values = []
        for mname, mvalue in kw.iteritems():
            aggregator = self._agg_from_name(mname)
            metric_name = strip_aggregator(
                self._get_full_metric_name(mname), aggregator)

            try:
                mvalue = float(mvalue)
            except (ValueError, TypeError):
                raise BadMetricsQueryError(
                    '%r is not a valid metric value,'
                    'should be a floating point number' % mvalue)

            metrics.append((Metric(metric_name, [aggregator]), mvalue))
            metrics_values.append({
                'name': mname,
                'value': mvalue,
                'aggregator': aggregator.name,
            })

        mm = yield self.backend.worker.get_metric_manager()
        [mm.oneshot(metric, value) for metric, value in metrics]

        returnValue(metrics_values)


class MetricWorker(Worker):
    def __init__(self, *args, **kwargs):
        super(MetricWorker, self).__init__(*args, **kwargs)
        self._started_d = Deferred()

    @inlineCallbacks
    def startWorker(self):
        self._metric_manager = yield self.start_publisher(
            MetricManager, '%s.' % self.config.get('prefix'))
        self._started_d.callback(self._metric_manager)

    @inlineCallbacks
    def stopWorker(self):
        if not self._started_d.called:
            yield self._started_d
        self._metric_manager.stop()

    def get_metric_manager(self):
        if self._started_d.called:
            return succeed(self._metric_manager)
        return self._started_d


class GraphiteBackendConfig(MetricsBackend.config_class):
    graphite_url = ConfigText(
        "Url for the graphite web server to query",
        default='http://127.0.0.1:8080')

    prefix = ConfigText(
        "Prefix for all metric names. Defaults to 'go.campaigns'",
        default='go.campaigns')

    persistent = ConfigBool(
        ("Flag given to treq telling it whether to maintain a single "
         "connection for the requests made to graphite's web app."),
        default=True)

    username = ConfigText(
        "Basic auth username for authenticating requests to graphite.",
        required=False)

    password = ConfigText(
        "Basic auth password for authenticating requests to graphite.",
        required=False)

    max_response_size = ConfigInt(
        ("Maximum number of data points to return. If a request specifies a "
         "time range and interval that contains more data than this, it is "
         "rejected."),
        default=10000)

    amqp_hostname = ConfigText(
        "Hostname for where AMQP broker is located",
        default='127.0.0.1')

    amqp_port = ConfigInt(
        "Port to connect to the AMQP broker",
        default=5672)

    amqp_username = ConfigText(
        "Username to connect to the AMQP broker",
        default="guest")

    amqp_password = ConfigText(
        "Password to connect to the AMQP broker",
        default="guest")

    amqp_vhost = ConfigText(
        "Virtualhost for AMQP broker",
        default="/")

    amqp_spec = ConfigText(
        "Spec file for AMQP",
        default="amqp-spec-0-8.xml")

    def post_validate(self):
        auth = (self.username, self.password)
        exists = [x is not None for x in auth]

        if any(exists) and not all(exists):
            raise ConfigError(
                "Either both a username and password need to be given or "
                "neither for graphite backend config")


class GraphiteBackend(MetricsBackend):
    model_class = GraphiteMetrics
    config_class = GraphiteBackendConfig

    def initialize(self):
        config = self.config
        worker_creator = WorkerCreator({
            'hostname': config.amqp_hostname,
            'port': config.amqp_port,
            'username': config.amqp_username,
            'password': config.amqp_password,
            'vhost': config.amqp_vhost,
            'specfile': config.amqp_spec,
            })
        self.worker = worker_creator.create_worker_by_class(
            MetricWorker, {
                'prefix': config.prefix
            })



"""
Time period and interval parameter parsers for Graphite backend.
"""

from datetime import datetime, timedelta
import re


class TimeParserValueError(ValueError):
    """
    A value could not be parsed.
    """


INTERVAL_RE = re.compile(r'(?P<count>\d+)(?P<unit>.+)')

UNIT_VALUES = {}


def _set_unit_value(value, *names):
    for name in names:
        UNIT_VALUES[name] = value


_set_unit_value(1, "s", "second", "seconds")
_set_unit_value(60, "min", "mins", "minute", "minutes")
_set_unit_value(3600, "h", "hour", "hours")
_set_unit_value(86400, "d", "day", "days")
_set_unit_value(7 * 86400, "w", "week", "weeks")
_set_unit_value(30 * 86400, "mon", "month", "months")
_set_unit_value(365 * 86400, "y", "year", "years")


def interval_to_seconds(interval_str):
    """
    Parse a time interval specifier of the form "<count><unit>" into the
    number of seconds contained in the interval.

    NOTE: This is stricter than Graphite's parser, which accepts any string
          starting with the shortest prefix for a unit.
    """
    parts = INTERVAL_RE.match(interval_str)
    if parts is None:
        raise TimeParserValueError(
            "Invalid interval string: %r" % (interval_str,))
    count = int(parts.groupdict()["count"])
    unit = parts.groupdict()["unit"]
    unit_multiplier = UNIT_VALUES.get(unit)
    if unit_multiplier is None:
        raise TimeParserValueError(
            "Invalid interval string: %r" % (interval_str,))
    return count * unit_multiplier


def _call_or_raise(exc, func, *args, **kw):
    """
    Call func(*args, **kw) and catch any exceptions, raising exc instead.

    This exists to avoid a bunch of boilerplate try/except blocks in
    parse_absolute_time()
    """
    try:
        return func(*args, **kw)
    except:
        raise exc


def parse_absolute_time(time_str):
    """
    Parse a Graphite-compatible absolute time specifier into a datetime object.

    This accepts `HH:MM_YYYYMMDD` and `YYYYMMDD` formats as well as unix
    timestamps.
    """
    # Build an exception to pass to _call_or_raise()
    exc = TimeParserValueError("Invalid time string: %r" % (time_str,))

    if ":" in time_str:
        return _call_or_raise(exc, datetime.strptime, time_str, "%H:%M_%Y%m%d")
    elif time_str.isdigit():
        # This is the same test graphite uses to determine whether a string is
        # a unix timestamp or a `YYYYMMDD` string. It's important that we make
        # the same decisions as graphite because differences could let very
        # expensive requests slip through and potentially break either the API
        # or graphite.
        if len(time_str) == 8 and all([int(time_str[:4]) > 1900,
                                       int(time_str[4:6]) < 13,
                                       int(time_str[6:]) < 32]):
            return _call_or_raise(exc, datetime.strptime, time_str, "%Y%m%d")
        else:
            return _call_or_raise(
                exc, datetime.utcfromtimestamp, int(time_str))
    else:
        raise exc


def parse_time(time_str, now):
    """
    Parse a Graphite-compatible absolute or relative time specifier into a
    datetime object.

    NOTE: This is stricter than Graphite's parser and accepts a narrower
          variety of formats. Currently, only relative time specifiers are
          supported.
    """
    if time_str in ["now", "today"]:
        return now
    if time_str == "yesterday":
        return now - timedelta(days=1)
    if time_str == "tomorrow":
        return now + timedelta(days=1)
    if time_str.startswith("-"):
        interval = interval_to_seconds(time_str[1:])
        return now - timedelta(seconds=interval)
    return parse_absolute_time(time_str)



from confmodel import Config


class MetricsBackendError(Exception):
    """
    Raised when an error occurs whilst interacting with a metrics backend.
    """


class BadMetricsQueryError(Exception):
    """
    Raised when an error occurs because a bad query was given.
    """


class Metrics(object):
    """
    A model encapsulating how metric values are queried for a particular
    backend and owner id. Intended to be subclassed with backend-specific
    logic.
    """

    def __init__(self, backend, owner_id):
        self.backend = backend
        self.owner_id = owner_id
        self.initialize()

    def initialize(self):
        """
        Optionally override for backend-specific setup
        """

    def get(self, **kw):
        """
        Override with backend-specific logic for retrieving values
        """
        raise NotImplementedError()

    def fire(self, **kw):
        """
        Override with backend-specific logic for firing metric values
        """
        raise NotImplementedError()


class MetricsBackendConfig(Config):
    """"
    Config for a metrics backend typed. Intended to be subclassed for
    backend-specific configuration
    """


class MetricsBackend(object):
    """
    Ensapsulates machinery specific to a metrics backend.
    """
    model_class = Metrics
    config_class = MetricsBackendConfig

    def __init__(self, config=None):
        if config is None:
            config = {}
        self.config = self.config_class(config)
        self.initialize()

    def initialize(self):
        """
        Optionally override for backend-specific setup
        """

    def get_model(self, owner_id):
        """
        Creates a new instance of the backend's metric model from the given
        owner id.
        """
        return self.model_class(self, owner_id)



import json
from base64 import b64encode

from twisted.trial.unittest import TestCase
from twisted.internet.defer import inlineCallbacks, returnValue

from confmodel.errors import ConfigError

from go_api.cyclone.helpers import MockHttpServer

from go_metrics.metrics.base import MetricsBackendError, BadMetricsQueryError
from go_metrics.metrics.graphite import (
    GraphiteMetrics, GraphiteBackend, GraphiteBackendConfig, MetricWorker)

from vumi.tests.utils import VumiWorkerTestCase


class TestGraphiteMetrics(VumiWorkerTestCase):

    @inlineCallbacks
    def mk_graphite(self, handler=None):
        graphite = MockHttpServer(handler)
        yield graphite.start()

        self.addCleanup(graphite.stop)
        returnValue(graphite)

    @inlineCallbacks
    def mk_backend(self, **kw):
        kw.setdefault('persistent', False)
        backend = GraphiteBackend(kw)
        backend.worker = yield self.get_worker({
            'prefix': backend.config.prefix,
            }, MetricWorker)
        returnValue(backend)

    @inlineCallbacks
    def test_get_request(self):
        reqs = []

        def handler(req):
            reqs.append(req)
            return '{}'

        graphite = yield self.mk_graphite(handler)
        backend = yield self.mk_backend(graphite_url=graphite.url)
        metrics = GraphiteMetrics(backend, 'owner-1')

        yield metrics.get(**{
            'm': ['stores.a.b.last', 'stores.b.a.max'],
            'from': '-48h',
            'until': '-24h',
            'interval': '1day'
        })

        [req] = reqs

        self.assertTrue(req.uri.startswith('/render/?'))

        self.assertEqual(req.args, {
            'format': ['json'],
            'from': ['-48h'],
            'until': ['-24h'],
            'target': [
                "alias(summarize(go.campaigns.owner-1.stores.a.b.last,"
                " '1day', 'last', false), 'stores.a.b.last')",

                "alias(summarize(go.campaigns.owner-1.stores.b.a.max, "
                "'1day', 'max', false), 'stores.b.a.max')"],
        })

    @inlineCallbacks
    def test_get_one_request(self):
        reqs = []

        def handler(req):
            reqs.append(req)
            return '{}'

        graphite = yield self.mk_graphite(handler)
        backend = yield self.mk_backend(graphite_url=graphite.url)
        metrics = GraphiteMetrics(backend, 'owner-1')

        yield metrics.get(**{
            'm': ['stores.a.b.last'],
            'from': '-48h',
            'until': '-24h',
            'interval': '1day'
        })

        [req] = reqs

        self.assertTrue(req.uri.startswith('/render/?'))

        self.assertEqual(req.args, {
            'format': ['json'],
            'from': ['-48h'],
            'until': ['-24h'],
            'target': [
                "alias(summarize(go.campaigns.owner-1.stores.a.b.last,"
                " '1day', 'last', false), 'stores.a.b.last')"],
        })

    @inlineCallbacks
    def test_get_one_request_no_list(self):
        reqs = []

        def handler(req):
            reqs.append(req)
            return '{}'

        graphite = yield self.mk_graphite(handler)
        backend = yield self.mk_backend(graphite_url=graphite.url)
        metrics = GraphiteMetrics(backend, 'owner-1')

        yield metrics.get(**{
            'm': 'stores.a.b.last',
            'from': '-48h',
            'until': '-24h',
            'interval': '1day'
        })

        [req] = reqs

        self.assertTrue(req.uri.startswith('/render/?'))

        self.assertEqual(req.args, {
            'format': ['json'],
            'from': ['-48h'],
            'until': ['-24h'],
            'target': [
                "alias(summarize(go.campaigns.owner-1.stores.a.b.last,"
                " '1day', 'last', false), 'stores.a.b.last')"],
        })

    @inlineCallbacks
    def test_get_response(self):
        def handler(req):
            return json.dumps([{
                'target': 'stores.a.b.last',
                'datapoints': [
                    [5.0, 5695],
                    [10.0, 5700]]
            }, {
                'target': 'stores.b.a.max',
                'datapoints': [
                    [12.0, 3724],
                    [14.0, 3741]]
            }])

        graphite = yield self.mk_graphite(handler)
        backend = yield self.mk_backend(graphite_url=graphite.url)
        metrics = GraphiteMetrics(backend, 'owner-1')

        data = yield metrics.get(m=['stores.a.b.last', 'stores.b.a.max'])

        self.assertEqual(data, {
            'stores.a.b.last': [{
                'x': 5695000,
                'y': 5.0
            }, {
                'x': 5700000,
                'y': 10.0
            }],
            'stores.b.a.max': [{
                'x': 3724000,
                'y': 12.0
            }, {
                'x': 3741000,
                'y': 14.0
            }]
        })

    @inlineCallbacks
    def test_get_default_metrics(self):
        reqs = []

        def handler(req):
            reqs.append(req)
            return '{}'

        graphite = yield self.mk_graphite(handler)
        backend = yield self.mk_backend(graphite_url=graphite.url)
        metrics = GraphiteMetrics(backend, 'owner-1')

        yield metrics.get(**{
            'from': '-48h',
            'until': '-24h',
            'interval': '1day'
        })

        [req] = reqs
        self.assertTrue('target' not in req.args)

    @inlineCallbacks
    def test_reject_large_request(self):
        """
        Requests for excessive amounts of data are rejected.
        """
        backend = yield self.mk_backend()
        metrics = GraphiteMetrics(backend, 'owner-1')

        err = yield self.assertFailure(
            metrics.get(**{'from': '-1d', 'interval': '1s'}),
            BadMetricsQueryError)
        self.assertEqual(
            str(err),
            "86400 data points requested, maximum allowed is 10000")

        # "from" can be later than "until".
        err = yield self.assertFailure(
            metrics.get(**{'from': 'now', 'until': '-1d', 'interval': '1s'}),
            BadMetricsQueryError)
        self.assertEqual(
            str(err),
            "86400 data points requested, maximum allowed is 10000")

    @inlineCallbacks
    def test_reject_large_request_multiple_metrics(self):
        """
        Requests for excessive amounts of data are rejected, even if the
        results for individual metrics are within the limits.
        """
        reqs = []

        def handler(req):
            reqs.append(req)
            return '{}'

        graphite = yield self.mk_graphite(handler)
        backend = yield self.mk_backend(graphite_url=graphite.url)
        metrics = GraphiteMetrics(backend, 'owner-1')

        # Two metrics, 8640 points each.
        err = yield self.assertFailure(
            metrics.get(**{
                'm': ['a', 'b'],
                'from': '-1d',
                'interval': '10s',
            }),
            BadMetricsQueryError)
        self.assertEqual(
            str(err),
            "17280 data points requested, maximum allowed is 10000")

        # Only one metric.
        resp = yield metrics.get(**{'from': '-1d', 'interval': '10s'})
        self.assertEqual(resp, {})

    @inlineCallbacks
    def test_reject_large_request_with_config(self):
        """
        The excessive data rejection threshold is configurable.
        """
        reqs = []

        def handler(req):
            reqs.append(req)
            return '{}'

        graphite = yield self.mk_graphite(handler)
        backend = yield self.mk_backend(
            graphite_url=graphite.url, max_response_size=100000)
        metrics = GraphiteMetrics(backend, 'owner-1')

        # Too much data.
        err = yield self.assertFailure(
            metrics.get(**{'from': '-2d', 'interval': '1s'}),
            BadMetricsQueryError)
        self.assertEqual(
            str(err),
            "172800 data points requested, maximum allowed is 100000")

        # Not too much data.
        resp = yield metrics.get(**{'from': '-1d', 'interval': '1s'})
        self.assertEqual(resp, {})

    @inlineCallbacks
    def test_get_defaults(self):
        reqs = []

        def handler(req):
            reqs.append(req)
            return '{}'

        graphite = yield self.mk_graphite(handler)
        backend = yield self.mk_backend(graphite_url=graphite.url)
        metrics = GraphiteMetrics(backend, 'owner-1')

        yield metrics.get(m=['stores.a.b.last'])

        [req] = reqs
        self.assertEqual(req.args, {
            'format': ['json'],
            'from': ['-24h'],
            'until': ['-0s'],
            'target': [
                "alias(summarize(go.campaigns.owner-1.stores.a.b.last,"
                " '1hour', 'last', false), 'stores.a.b.last')"],
        })

    @inlineCallbacks
    def test_get_backend_error(self):
        def handler(req):
            req.setResponseCode(400)
            return ':('

        graphite = yield self.mk_graphite(handler)
        backend = yield self.mk_backend(graphite_url=graphite.url)
        metrics = GraphiteMetrics(backend, 'owner-1')

        err = yield self.assertFailure(metrics.get(), MetricsBackendError)
        self.assertEqual(
            str(err), "Got error response interacting with metrics backend")

    @inlineCallbacks
    def test_get_null_handling_default(self):
        def handler(req):
            return json.dumps([{
                'target': 'stores.a.b.last',
                'datapoints': [
                    [None, 2695],
                    [5.0, 3695],
                    [None, 4695],
                    [None, 5695],
                    [10.0, 6695],
                    [None, 7695]]
            }, {
                'target': 'stores.b.a.max',
                'datapoints': [
                    [None, 2695],
                    [12.0, 3695],
                    [None, 4695],
                    [14.0, 5695]]
            }])

        graphite = yield self.mk_graphite(handler)
        backend = yield self.mk_backend(graphite_url=graphite.url)
        metrics = GraphiteMetrics(backend, 'owner-1')

        self.assertEqual(
            (yield metrics.get(
                m=['stores.a.b.last', 'stores.b.a.max'])),
            (yield metrics.get(
                m=['stores.a.b.last', 'stores.b.a.max'],
                nulls='zeroize')))

    @inlineCallbacks
    def test_get_null_handling_zeroize(self):
        def handler(req):
            return json.dumps([{
                'target': 'stores.a.b.last',
                'datapoints': [
                    [None, 2695],
                    [5.0, 3695],
                    [None, 4695],
                    [None, 5695],
                    [10.0, 6695],
                    [None, 7695]]
            }, {
                'target': 'stores.b.a.max',
                'datapoints': [
                    [None, 2695],
                    [12.0, 3695],
                    [None, 4695],
                    [14.0, 5695]]
            }])

        graphite = yield self.mk_graphite(handler)
        backend = yield self.mk_backend(graphite_url=graphite.url)
        metrics = GraphiteMetrics(backend, 'owner-1')

        data = yield metrics.get(
            m=['stores.a.b.last', 'stores.b.a.max'],
            nulls='zeroize')

        self.assertEqual(data, {
            'stores.a.b.last': [{
                'x': 2695000,
                'y': 0.0
            }, {
                'x': 3695000,
                'y': 5.0
            }, {
                'x': 4695000,
                'y': 0.0
            }, {
                'x': 5695000,
                'y': 0.0
            }, {
                'x': 6695000,
                'y': 10.0
            }, {
                'x': 7695000,
                'y': 0.0
            }],
            'stores.b.a.max': [{
                'x': 2695000,
                'y': 0.0
            }, {
                'x': 3695000,
                'y': 12.0
            }, {
                'x': 4695000,
                'y': 0.0
            }, {
                'x': 5695000,
                'y': 14.0
            }]
        })

    @inlineCallbacks
    def test_get_null_handling_omit(self):
        def handler(req):
            return json.dumps([{
                'target': 'stores.a.b.last',
                'datapoints': [
                    [None, 2695],
                    [5.0, 3695],
                    [None, 4695],
                    [None, 5695],
                    [10.0, 6695],
                    [None, 7695]]
            }, {
                'target': 'stores.b.a.max',
                'datapoints': [
                    [None, 2695],
                    [12.0, 3695],
                    [None, 4695],
                    [14.0, 5695]]
            }])

        graphite = yield self.mk_graphite(handler)
        backend = yield self.mk_backend(graphite_url=graphite.url)
        metrics = GraphiteMetrics(backend, 'owner-1')

        data = yield metrics.get(
            m=['stores.a.b.last', 'stores.b.a.max'],
            nulls='omit')

        self.assertEqual(data, {
            'stores.a.b.last': [{
                'x': 3695000,
                'y': 5.0
            }, {
                'x': 6695000,
                'y': 10.0
            }],
            'stores.b.a.max': [{
                'x': 3695000,
                'y': 12.0
            }, {
                'x': 5695000,
                'y': 14.0
            }]
        })

    @inlineCallbacks
    def test_get_null_handling_keep(self):
        def handler(req):
            return json.dumps([{
                'target': 'stores.a.b.last',
                'datapoints': [
                    [None, 2695],
                    [5.0, 3695],
                    [None, 4695],
                    [None, 5695],
                    [10.0, 6695],
                    [None, 7695]]
            }, {
                'target': 'stores.b.a.max',
                'datapoints': [
                    [None, 2695],
                    [12.0, 3695],
                    [None, 4695],
                    [14.0, 5695]]
            }])

        graphite = yield self.mk_graphite(handler)
        backend = yield self.mk_backend(graphite_url=graphite.url)
        metrics = GraphiteMetrics(backend, 'owner-1')

        data = yield metrics.get(
            m=['stores.a.b.last', 'stores.b.a.max'],
            nulls='keep')

        self.assertEqual(data, {
            'stores.a.b.last': [{
                'x': 2695000,
                'y': None
            }, {
                'x': 3695000,
                'y': 5.0
            }, {
                'x': 4695000,
                'y': None
            }, {
                'x': 5695000,
                'y': None
            }, {
                'x': 6695000,
                'y': 10.0
            }, {
                'x': 7695000,
                'y': None
            }],
            'stores.b.a.max': [{
                'x': 2695000,
                'y': None
            }, {
                'x': 3695000,
                'y': 12.0
            }, {
                'x': 4695000,
                'y': None
            }, {
                'x': 5695000,
                'y': 14.0
            }]
        })

    @inlineCallbacks
    def test_get_null_handling_unrecognised(self):
        graphite = yield self.mk_graphite()
        backend = yield self.mk_backend(graphite_url=graphite.url)
        metrics = GraphiteMetrics(backend, 'owner-1')

        err = yield self.assertFailure(
            metrics.get(m=['stores.a.b.last', 'stores.b.a.max'], nulls='bad'),
            BadMetricsQueryError)
        self.assertEqual(str(err), "Unrecognised null parser 'bad'")

    @inlineCallbacks
    def test_get_auth(self):
        reqs = []

        def handler(req):
            reqs.append(req)
            return '{}'

        graphite = yield self.mk_graphite(handler)
        backend = yield self.mk_backend(
            graphite_url=graphite.url,
            username="root",
            password="toor")

        metrics = GraphiteMetrics(backend, 'owner-1')

        yield metrics.get(**{
            'm': ['stores.a.b.last', 'stores.b.a.max'],
            'from': '-48h',
            'until': '-24h',
            'interval': '1day'
        })

        [req] = reqs
        self.assertEqual(
            req.getHeader('Authorization'),
            'Basic %s' % (b64encode('root:toor')))

    @inlineCallbacks
    def test_get_no_auth(self):
        reqs = []

        def handler(req):
            reqs.append(req)
            return '{}'

        graphite = yield self.mk_graphite(handler)
        backend = yield self.mk_backend(graphite_url=graphite.url)

        metrics = GraphiteMetrics(backend, 'owner-1')

        yield metrics.get(**{
            'm': ['stores.a.b.last', 'stores.b.a.max'],
            'from': '-48h',
            'until': '-24h',
            'interval': '1day'
        })

        [req] = reqs
        self.assertEqual(req.getHeader('Authorization'), None)

    @inlineCallbacks
    def test_post_request_single(self):
        backend = yield self.mk_backend()
        metrics = GraphiteMetrics(backend, 'owner-1')

        res = yield metrics.fire(**{'foo.avg': 1.7})
        [metric] = res

        self.assertEqual(metric, {
            'name': 'foo.avg',
            'value': 1.7,
            'aggregator': 'avg',
        })

    @inlineCallbacks
    def test_post_request_multiple(self):
        backend = yield self.mk_backend()
        metrics = GraphiteMetrics(backend, 'owner-1')

        res = yield metrics.fire(**{'foo.avg': 1.2, 'oof.avg': 2.7})

        self.assertEqual(sorted(res), sorted([{
            'name': 'foo.avg',
            'value': 1.2,
            'aggregator': 'avg',
        }, {
            'name': 'oof.avg',
            'value': 2.7,
            'aggregator': 'avg',
        }]))

    @inlineCallbacks
    def test_post_request_bad_value(self):
        backend = yield self.mk_backend()
        metrics = GraphiteMetrics(backend, 'owner-1')

        try:
            yield metrics.fire(**{'foo.avg': 'bar'})
        except BadMetricsQueryError as e:
            self.assertEqual(
                str(e), "'bar' is not a valid metric value,should be a "
                "floating point number")
        else:
            self.fail("BadMetricQueryError not raised")

    @inlineCallbacks
    def test_post_request_bad_aggregator(self):
        backend = yield self.mk_backend()
        metrics = GraphiteMetrics(backend, 'owner-1')

        try:
            yield metrics.fire(**{'foo': 1.2})
        except BadMetricsQueryError as e:
            self.assertEqual(
                str(e), "Aggregator 'foo' is not a valid aggregator")
        else:
            self.fail("BadMetricQueryError not raised")

    @inlineCallbacks
    def test_post_aggregators(self):
        backend = yield self.mk_backend()
        metrics = GraphiteMetrics(backend, 'owner-1')

        for agg in GraphiteMetrics.aggregators:
            res = yield metrics.fire(**{'foo.%s' % agg: 1.3})
            [metric] = res
            self.assertEqual(metric['aggregator'], agg)


class TestGraphiteBackendConfig(TestCase):
    def test_auth_fields(self):
        GraphiteBackendConfig({})
        GraphiteBackendConfig({
            'username': 'foo',
            'password': 'bar',
        })
        self.assertRaises(
            ConfigError, GraphiteBackendConfig, {'username': 'foo'})
        self.assertRaises(
            ConfigError, GraphiteBackendConfig, {'password': 'bar'})



"""
Tests for the metrics API's server.
"""
import json
import yaml

from twisted.internet.defer import succeed, inlineCallbacks, maybeDeferred
from twisted.trial.unittest import TestCase

from confmodel.fields import ConfigText

from go_api.cyclone.helpers import AppHelper

from go_metrics.server import MetricsApi
from go_metrics.metrics.base import MetricsBackendError, BadMetricsQueryError
from go_metrics.metrics.dummy import DummyBackend


class DummyMetricsApi(MetricsApi):
    config_required = False
    backend_class = DummyBackend
    factory_preprocessor = staticmethod(lambda x: x)


class TestMetricsApi(TestCase):
    def mk_config(self, **kw):
        tempfile = self.mktemp()
        with open(tempfile, 'wb') as fp:
            yaml.safe_dump(kw, fp)

        return tempfile

    def test_initialize_backend(self):
        class ToyBackendConfig(DummyBackend.config_class):
            foo = ConfigText("A foo")

        class ToyBackend(DummyBackend):
            config_class = ToyBackendConfig

        class ToyApi(DummyMetricsApi):
            backend_class = ToyBackend

        app = ToyApi(self.mk_config(backend={'foo': 'bar'}))
        self.assertTrue(isinstance(app.backend, ToyBackend))
        self.assertTrue(isinstance(app.backend.config, ToyBackendConfig))
        self.assertEqual(app.backend.config.foo, 'bar')

    def test_get_metrics_model(self):
        app = DummyMetricsApi(self.mk_config())
        model = app.get_metrics_model('owner-1')
        self.assertEqual(model.owner_id, 'owner-1')

    @inlineCallbacks
    def test_metrics_get(self):
        app = DummyMetricsApi(self.mk_config())
        app.backend.fixtures.add(
            foo='bar',
            baz=['quux', 'corge'],
            result={'grault': 'garply'})

        get = AppHelper(app).get
        resp = yield get('/metrics/', params={
            'foo': 'bar',
            'baz': ['quux', 'corge']
        })
        self.assertEqual((yield resp.json()), {'grault': 'garply'})

    @inlineCallbacks
    def test_metrics_get_async(self):
        app = DummyMetricsApi(self.mk_config())
        app.backend.fixtures.add(foo='bar', result=succeed({'baz': 'quux'}))
        get = AppHelper(app).get
        resp = yield get('/metrics/', params={'foo': 'bar'})
        self.assertEqual((yield resp.json()), {'baz': 'quux'})

    @inlineCallbacks
    def test_metrics_get_query_error(self):
        app = DummyMetricsApi(self.mk_config())

        def fail():
            raise BadMetricsQueryError(":(")

        app.backend.fixtures.add(foo='bar', result=maybeDeferred(fail))

        get = AppHelper(app).get
        resp = yield get('/metrics/', params={'foo': 'bar'})

        self.assertEqual((yield resp.json()), {
            'status_code': 400,
            'reason': ':(',
        })

    @inlineCallbacks
    def test_metrics_get_backend_error(self):
        app = DummyMetricsApi(self.mk_config())

        def fail():
            raise MetricsBackendError(":(")

        app.backend.fixtures.add(foo='bar', result=maybeDeferred(fail))

        get = AppHelper(app).get
        resp = yield get('/metrics/', params={'foo': 'bar'})

        self.assertEqual((yield resp.json()), {
            'status_code': 500,
            'reason': ':(',
        })

    @inlineCallbacks
    def test_metrics_get_uncaught_error(self):
        app = DummyMetricsApi(self.mk_config())

        class DummyError(Exception):
            pass

        def fail():
            raise DummyError(":(")

        app.backend.fixtures.add(foo='bar', result=maybeDeferred(fail))

        get = AppHelper(app).get
        resp = yield get('/metrics/', params={'foo': 'bar'})

        self.assertEqual((yield resp.json()), {
            'status_code': 500,
            'reason': 'Failed to retrieve metrics.',
        })

        [f] = self.flushLoggedErrors(DummyError)
        self.assertEqual(str(f.value), ":(")

    @inlineCallbacks
    def test_metrics_post(self):
        app = DummyMetricsApi(self.mk_config())
        app.backend.fixtures.add(
            foo='bar',
            baz=['quux', 'corge'],
            method='fire',
            result={'grault': 'garply'})

        post = AppHelper(app).post
        resp = yield post('/metrics/', data=json.dumps({
            'foo': 'bar',
            'baz': ['quux', 'corge']
        }))
        self.assertEqual((yield resp.json()), {'grault': 'garply'})

    @inlineCallbacks
    def test_metrics_post_async(self):
        app = DummyMetricsApi(self.mk_config())
        app.backend.fixtures.add(
            foo='bar', result=succeed({'baz': 'quux'}), method='fire')
        post = AppHelper(app).post
        resp = yield post('/metrics/', data=json.dumps({'foo': 'bar'}))
        self.assertEqual((yield resp.json()), {'baz': 'quux'})

    @inlineCallbacks
    def test_metrics_post_query_error(self):
        app = DummyMetricsApi(self.mk_config())

        def fail():
            raise BadMetricsQueryError(":(")

        app.backend.fixtures.add(
            foo='bar', result=maybeDeferred(fail), method='fire')

        post = AppHelper(app).post
        resp = yield post('/metrics/', data=json.dumps({'foo': 'bar'}))

        self.assertEqual((yield resp.json()), {
            'status_code': 400,
            'reason': ':(',
        })

    @inlineCallbacks
    def test_metrics_post_uncaught_error(self):
        app = DummyMetricsApi(self.mk_config())

        class DummyError(Exception):
            pass

        def fail():
            raise DummyError(":(")

        app.backend.fixtures.add(
            foo='bar', result=maybeDeferred(fail), method='fire')

        post = AppHelper(app).post
        resp = yield post('/metrics/', data=json.dumps({'foo': 'bar'}))

        self.assertEqual((yield resp.json()), {
            'status_code': 500,
            'reason': 'Failed to fire metrics.',
        })

        [f] = self.flushLoggedErrors(DummyError)
        self.assertEqual(str(f.value), ":(")



from twisted.trial.unittest import TestCase

from go_metrics.metrics.dummy import Fixtures, DummyMetrics, DummyBackend


class TestFixtures(TestCase):
    def test_add(self):
        fixtures = Fixtures()
        fixtures.add(foo='bar', method='fire', result={'baz': 'quux'})
        self.assertEqual(fixtures.items, [{
            'foo': 'bar',
            'method': 'fire',
            'result': {'baz': 'quux'}
        }])

    def test_add_defaults(self):
        fixtures = Fixtures()
        fixtures.add(foo='bar')
        self.assertEqual(fixtures.items, [{
            'foo': 'bar',
            'method': 'get',
            'result': {}
        }])

    def test_match(self):
        fixtures = Fixtures()
        fixtures.add(foo='bar', result={'baz': 'quux'})
        fixtures.add(corge='grault', result={'garply': 'waldo'})
        self.assertEqual(fixtures.match(foo='bar'), {'baz': 'quux'})
        self.assertEqual(fixtures.match(corge='grault'), {'garply': 'waldo'})
        self.assertEqual(fixtures.match(fred='xzyy'), {})


class TestDummyMetrics(TestCase):
    def test_get(self):
        backend = DummyBackend()
        metrics = DummyMetrics(backend, 'owner-1')
        backend.fixtures.add(foo='bar', result={'baz': 'quux'})
        self.assertEqual(metrics.get(foo='bar'), {'baz': 'quux'})


class TestDummyBackend(TestCase):
    pass



from setuptools import setup, find_packages

setup(
    name="go-metrics-api",
    version="0.1.5",
    url='http://github.com/praekelt/go-metrics-api',
    license='BSD',
    description="An API for querying vumi-go metrics",
    long_description=open('README.rst', 'r').read(),
    author='Praekelt Foundation',
    author_email='dev@praekeltfoundation.org',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'treq',
        'cyclone',
        'go_api>=0.2.0',
        'confmodel==0.2.0',
        'vumi>=0.5.21',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet :: WWW/HTTP',
    ],
)



"""
Dummy backend for the metrics api for use in tests.
"""

from go_metrics.metrics.base import Metrics, MetricsBackend


def matches(target, items):
    for k, v in items.iteritems():
        if k not in target:
            return False

        if target[k] != items[k]:
            return False

    return True


class Fixtures(object):
    def __init__(self):
        self.items = []

    def match(self, **kw):
        for f in self.items:
            if matches(f, kw):
                return f['result']
        return {}

    def add(self, **kw):
        kw.setdefault('method', 'get')
        kw.setdefault('result', {})
        self.items.append(kw)


class DummyMetrics(Metrics):
    def get(self, **kw):
        return self.backend.fixtures.match(method='get', **kw)

    def fire(self, **kw):
        return self.backend.fixtures.match(method='fire', **kw)


class DummyBackend(MetricsBackend):
    model_class = DummyMetrics

    def initialize(self):
        self.fixtures = Fixtures()
